import re
import pandas as pd
import numpy as np
from scipy.stats import entropy
from itertools import chain
import glob
from random import choices

import matplotlib.pyplot as plt
import seaborn as sns

REGEX = re.compile(
    r"""^
    (\.)?
    ((?P<key>[a-gA-G](b*|\#*)|(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i))\.)?
    ((?P<pedal>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i))\[)?
    (?P<numeral>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i|Ger|It|Fr|@none))
    (?P<form>[%o+M])?
    (?P<figbass>(9|7|65|43|42|2|64|6))?
    (\((?P<changes>(\+?(b*|\#*)\d)+)\))?
    (/\.?(?P<relativeroot>(b*|\#*)(VII|VI|V|IV|III|II|I|vii|vi|v|iv|iii|ii|i)))?
    (?P<pedalend>\])?
    (?P<phraseend>\\\\)?$
""", re.X
)

def grams(arr, n):
    """Returns a list of n-grams for given array."""
    assert n>0
    arr = np.asarray(arr)
    l = len(arr)

    # get n-grams
    grams = [ arr[i:(i+n)] for i in range(l-n+1) ]
    # convert to tuple of strings
    grams = [tuple(str(g) for g in gram) for gram in grams]

    return grams

def normalize_df(df, ax=None):
    # normalize by sum
    if not ax:
        return df / df.sum(axis=1).sum(axis=0)
    # normalize columns
    if ax==0:
        return df.div(df.sum(axis=0), axis=1)
    # normalize rows
    elif ax==1:
        return df.div(df.sum(axis=1), axis=0)
    else:
        print('Axis for normalization not specified')

def normalized_entropy(counts):
    """Calculate normalized entropy."""
    support = [c for c in counts if c > 0 ]
    if len(support) == 1:
        H = 0
    else:
        H = entropy(counts, base=2) / np.log2(len(support))
    return H

def root_distance_asc(first_chord, second_chord):
    """Calculate diatonic root distance and direction between two chords."""
    # the 12th group is the numeral without accidentals
    first_root = re.match(REGEX, first_chord).groups()[12].upper()
    second_root = re.match(REGEX, second_chord).groups()[12].upper()

    sd_dict = {
        'I':1,
        'II':2,
        'III':3,
        'IV':4,
        'V':5,
        'VI':6,
        'VII':7,
        'GER':6,
        'FR':6,
        'IT':6
    }

    distance = (sd_dict[second_root] - sd_dict[first_root]) % 7 #+ 1
    direction = np.sign(distance)
    distance += direction

    return distance

def onesamp_bootstrap(samples, mu_0, N_B=100_000):
    """Perform one-sample boostrap hypothesis test."""

    # make sample conform to null hypothesis
    mu = np.mean(samples)
    shifted_samples = [ s - mu + mu_0 for s in samples]

    # resampling
    resampled_means = [ np.mean(choices(shifted_samples, k=len(shifted_samples))) for i in range(N_B)]

    # p-value

    p = 2/N_B * min(
                    len([r for r in resampled_means if r <= mu]),
                    len([r for r in resampled_means if r >= mu])
                    )

    return resampled_means, p
